# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hangar_service.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hangar_service.proto',
  package='hangar',
  syntax='proto3',
  serialized_options=_b('H\001'),
  serialized_pb=_b('\n\x14hangar_service.proto\x12\x06hangar\"+\n\nErrorProto\x12\x0c\n\x04\x63ode\x18\x01 \x01(\x03\x12\x0f\n\x07message\x18\x02 \x01(\t\",\n\x0c\x42ranchRecord\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0e\n\x06\x63ommit\x18\x02 \x01(\t\"*\n\nHashRecord\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x0e\n\x06\x64igest\x18\x02 \x01(\t\"9\n\x0c\x43ommitRecord\x12\x0e\n\x06parent\x18\x01 \x01(\x0c\x12\x0b\n\x03ref\x18\x02 \x01(\x0c\x12\x0c\n\x04spec\x18\x03 \x01(\x0c\",\n\x0cSchemaRecord\x12\x0e\n\x06\x64igest\x18\x01 \x01(\t\x12\x0c\n\x04\x62lob\x18\x02 \x01(\x0c\"\r\n\x0bPingRequest\"\x1b\n\tPingReply\x12\x0e\n\x06result\x18\x01 \x01(\t\"\x18\n\x16GetClientConfigRequest\"\xa2\x01\n\x14GetClientConfigReply\x12\x38\n\x06\x63onfig\x18\x01 \x03(\x0b\x32(.hangar.GetClientConfigReply.ConfigEntry\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\x1a-\n\x0b\x43onfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"=\n\x18\x46\x65tchBranchRecordRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\"^\n\x16\x46\x65tchBranchRecordReply\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\"s\n\x10\x46\x65tchDataRequest\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"q\n\x0e\x46\x65tchDataReply\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"$\n\x12\x46\x65tchCommitRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\"\x84\x01\n\x10\x46\x65tchCommitReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x17\n\x0ftotal_byte_size\x18\x02 \x01(\x03\x12$\n\x06record\x18\x03 \x01(\x0b\x32\x14.hangar.CommitRecord\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"7\n\x12\x46\x65tchSchemaRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\"X\n\x10\x46\x65tchSchemaReply\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\x12!\n\x05\x65rror\x18\x02 \x01(\x0b\x32\x12.hangar.ErrorProto\"<\n\x17PushBranchRecordRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.BranchRecord\":\n\x15PushBranchRecordReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"r\n\x0fPushDataRequest\x12\x10\n\x08raw_data\x18\x01 \x01(\x0c\x12\x13\n\x0b\x63omp_nbytes\x18\x02 \x01(\x03\x12\x15\n\runcomp_nbytes\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"2\n\rPushDataReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"b\n\x11PushCommitRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x17\n\x0ftotal_byte_size\x18\x02 \x01(\x03\x12$\n\x06record\x18\x03 \x01(\x0b\x32\x14.hangar.CommitRecord\"4\n\x0fPushCommitReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"6\n\x11PushSchemaRequest\x12!\n\x03rec\x18\x01 \x01(\x0b\x32\x14.hangar.SchemaRecord\"4\n\x0fPushSchemaReply\x12!\n\x05\x65rror\x18\x01 \x01(\x0b\x32\x12.hangar.ErrorProto\"R\n\x19\x46indMissingCommitsRequest\x12\x0f\n\x07\x63ommits\x18\x01 \x03(\t\x12$\n\x06\x62ranch\x18\x02 \x01(\x0b\x32\x14.hangar.BranchRecord\"s\n\x17\x46indMissingCommitsReply\x12\x0f\n\x07\x63ommits\x18\x01 \x03(\t\x12$\n\x06\x62ranch\x18\x02 \x01(\x0b\x32\x14.hangar.BranchRecord\x12!\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x12.hangar.ErrorProto\"W\n\x1d\x46indMissingHashRecordsRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\r\n\x05hashs\x18\x02 \x01(\x0c\x12\x17\n\x0ftotal_byte_size\x18\x03 \x01(\x03\"x\n\x1b\x46indMissingHashRecordsReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\r\n\x05hashs\x18\x02 \x01(\x0c\x12\x17\n\x0ftotal_byte_size\x18\x03 \x01(\x03\x12!\n\x05\x65rror\x18\x04 \x01(\x0b\x32\x12.hangar.ErrorProto\"C\n\x19\x46indMissingSchemasRequest\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x16\n\x0eschema_digests\x18\x02 \x03(\t\"d\n\x17\x46indMissingSchemasReply\x12\x0e\n\x06\x63ommit\x18\x01 \x01(\t\x12\x16\n\x0eschema_digests\x18\x02 \x03(\t\x12!\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x12.hangar.ErrorProto2\xc5\n\n\rHangarService\x12\x30\n\x04PING\x12\x13.hangar.PingRequest\x1a\x11.hangar.PingReply\"\x00\x12Q\n\x0fGetClientConfig\x12\x1e.hangar.GetClientConfigRequest\x1a\x1c.hangar.GetClientConfigReply\"\x00\x12W\n\x11\x46\x65tchBranchRecord\x12 .hangar.FetchBranchRecordRequest\x1a\x1e.hangar.FetchBranchRecordReply\"\x00\x12\x43\n\tFetchData\x12\x18.hangar.FetchDataRequest\x1a\x16.hangar.FetchDataReply\"\x00(\x01\x30\x01\x12G\n\x0b\x46\x65tchCommit\x12\x1a.hangar.FetchCommitRequest\x1a\x18.hangar.FetchCommitReply\"\x00\x30\x01\x12\x45\n\x0b\x46\x65tchSchema\x12\x1a.hangar.FetchSchemaRequest\x1a\x18.hangar.FetchSchemaReply\"\x00\x12T\n\x10PushBranchRecord\x12\x1f.hangar.PushBranchRecordRequest\x1a\x1d.hangar.PushBranchRecordReply\"\x00\x12>\n\x08PushData\x12\x17.hangar.PushDataRequest\x1a\x15.hangar.PushDataReply\"\x00(\x01\x12\x44\n\nPushCommit\x12\x19.hangar.PushCommitRequest\x1a\x17.hangar.PushCommitReply\"\x00(\x01\x12\x42\n\nPushSchema\x12\x19.hangar.PushSchemaRequest\x1a\x17.hangar.PushSchemaReply\"\x00\x12_\n\x17\x46\x65tchFindMissingCommits\x12!.hangar.FindMissingCommitsRequest\x1a\x1f.hangar.FindMissingCommitsReply\"\x00\x12o\n\x1b\x46\x65tchFindMissingHashRecords\x12%.hangar.FindMissingHashRecordsRequest\x1a#.hangar.FindMissingHashRecordsReply\"\x00(\x01\x30\x01\x12_\n\x17\x46\x65tchFindMissingSchemas\x12!.hangar.FindMissingSchemasRequest\x1a\x1f.hangar.FindMissingSchemasReply\"\x00\x12^\n\x16PushFindMissingCommits\x12!.hangar.FindMissingCommitsRequest\x1a\x1f.hangar.FindMissingCommitsReply\"\x00\x12n\n\x1aPushFindMissingHashRecords\x12%.hangar.FindMissingHashRecordsRequest\x1a#.hangar.FindMissingHashRecordsReply\"\x00(\x01\x30\x01\x12^\n\x16PushFindMissingSchemas\x12!.hangar.FindMissingSchemasRequest\x1a\x1f.hangar.FindMissingSchemasReply\"\x00\x42\x02H\x01\x62\x06proto3')
)




_ERRORPROTO = _descriptor.Descriptor(
  name='ErrorProto',
  full_name='hangar.ErrorProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='hangar.ErrorProto.code', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='hangar.ErrorProto.message', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=32,
  serialized_end=75,
)


_BRANCHRECORD = _descriptor.Descriptor(
  name='BranchRecord',
  full_name='hangar.BranchRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='hangar.BranchRecord.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.BranchRecord.commit', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=77,
  serialized_end=121,
)


_HASHRECORD = _descriptor.Descriptor(
  name='HashRecord',
  full_name='hangar.HashRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='type', full_name='hangar.HashRecord.type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='digest', full_name='hangar.HashRecord.digest', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=123,
  serialized_end=165,
)


_COMMITRECORD = _descriptor.Descriptor(
  name='CommitRecord',
  full_name='hangar.CommitRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='parent', full_name='hangar.CommitRecord.parent', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ref', full_name='hangar.CommitRecord.ref', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='spec', full_name='hangar.CommitRecord.spec', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=167,
  serialized_end=224,
)


_SCHEMARECORD = _descriptor.Descriptor(
  name='SchemaRecord',
  full_name='hangar.SchemaRecord',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='digest', full_name='hangar.SchemaRecord.digest', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='blob', full_name='hangar.SchemaRecord.blob', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=226,
  serialized_end=270,
)


_PINGREQUEST = _descriptor.Descriptor(
  name='PingRequest',
  full_name='hangar.PingRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=285,
)


_PINGREPLY = _descriptor.Descriptor(
  name='PingReply',
  full_name='hangar.PingReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='hangar.PingReply.result', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=287,
  serialized_end=314,
)


_GETCLIENTCONFIGREQUEST = _descriptor.Descriptor(
  name='GetClientConfigRequest',
  full_name='hangar.GetClientConfigRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=316,
  serialized_end=340,
)


_GETCLIENTCONFIGREPLY_CONFIGENTRY = _descriptor.Descriptor(
  name='ConfigEntry',
  full_name='hangar.GetClientConfigReply.ConfigEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='hangar.GetClientConfigReply.ConfigEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='hangar.GetClientConfigReply.ConfigEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=505,
)

_GETCLIENTCONFIGREPLY = _descriptor.Descriptor(
  name='GetClientConfigReply',
  full_name='hangar.GetClientConfigReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='config', full_name='hangar.GetClientConfigReply.config', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.GetClientConfigReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GETCLIENTCONFIGREPLY_CONFIGENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=343,
  serialized_end=505,
)


_FETCHBRANCHRECORDREQUEST = _descriptor.Descriptor(
  name='FetchBranchRecordRequest',
  full_name='hangar.FetchBranchRecordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchBranchRecordRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=507,
  serialized_end=568,
)


_FETCHBRANCHRECORDREPLY = _descriptor.Descriptor(
  name='FetchBranchRecordReply',
  full_name='hangar.FetchBranchRecordReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchBranchRecordReply.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchBranchRecordReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=570,
  serialized_end=664,
)


_FETCHDATAREQUEST = _descriptor.Descriptor(
  name='FetchDataRequest',
  full_name='hangar.FetchDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.FetchDataRequest.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.FetchDataRequest.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.FetchDataRequest.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchDataRequest.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=666,
  serialized_end=781,
)


_FETCHDATAREPLY = _descriptor.Descriptor(
  name='FetchDataReply',
  full_name='hangar.FetchDataReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.FetchDataReply.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.FetchDataReply.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.FetchDataReply.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchDataReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=783,
  serialized_end=896,
)


_FETCHCOMMITREQUEST = _descriptor.Descriptor(
  name='FetchCommitRequest',
  full_name='hangar.FetchCommitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FetchCommitRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=898,
  serialized_end=934,
)


_FETCHCOMMITREPLY = _descriptor.Descriptor(
  name='FetchCommitReply',
  full_name='hangar.FetchCommitReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FetchCommitReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FetchCommitReply.total_byte_size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='record', full_name='hangar.FetchCommitReply.record', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchCommitReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=937,
  serialized_end=1069,
)


_FETCHSCHEMAREQUEST = _descriptor.Descriptor(
  name='FetchSchemaRequest',
  full_name='hangar.FetchSchemaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchSchemaRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1071,
  serialized_end=1126,
)


_FETCHSCHEMAREPLY = _descriptor.Descriptor(
  name='FetchSchemaReply',
  full_name='hangar.FetchSchemaReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.FetchSchemaReply.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FetchSchemaReply.error', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1128,
  serialized_end=1216,
)


_PUSHBRANCHRECORDREQUEST = _descriptor.Descriptor(
  name='PushBranchRecordRequest',
  full_name='hangar.PushBranchRecordRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.PushBranchRecordRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1218,
  serialized_end=1278,
)


_PUSHBRANCHRECORDREPLY = _descriptor.Descriptor(
  name='PushBranchRecordReply',
  full_name='hangar.PushBranchRecordReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushBranchRecordReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1280,
  serialized_end=1338,
)


_PUSHDATAREQUEST = _descriptor.Descriptor(
  name='PushDataRequest',
  full_name='hangar.PushDataRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='raw_data', full_name='hangar.PushDataRequest.raw_data', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='comp_nbytes', full_name='hangar.PushDataRequest.comp_nbytes', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uncomp_nbytes', full_name='hangar.PushDataRequest.uncomp_nbytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushDataRequest.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1340,
  serialized_end=1454,
)


_PUSHDATAREPLY = _descriptor.Descriptor(
  name='PushDataReply',
  full_name='hangar.PushDataReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushDataReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1456,
  serialized_end=1506,
)


_PUSHCOMMITREQUEST = _descriptor.Descriptor(
  name='PushCommitRequest',
  full_name='hangar.PushCommitRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.PushCommitRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.PushCommitRequest.total_byte_size', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='record', full_name='hangar.PushCommitRequest.record', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1508,
  serialized_end=1606,
)


_PUSHCOMMITREPLY = _descriptor.Descriptor(
  name='PushCommitReply',
  full_name='hangar.PushCommitReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushCommitReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1608,
  serialized_end=1660,
)


_PUSHSCHEMAREQUEST = _descriptor.Descriptor(
  name='PushSchemaRequest',
  full_name='hangar.PushSchemaRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='rec', full_name='hangar.PushSchemaRequest.rec', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1662,
  serialized_end=1716,
)


_PUSHSCHEMAREPLY = _descriptor.Descriptor(
  name='PushSchemaReply',
  full_name='hangar.PushSchemaReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.PushSchemaReply.error', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1718,
  serialized_end=1770,
)


_FINDMISSINGCOMMITSREQUEST = _descriptor.Descriptor(
  name='FindMissingCommitsRequest',
  full_name='hangar.FindMissingCommitsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commits', full_name='hangar.FindMissingCommitsRequest.commits', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='branch', full_name='hangar.FindMissingCommitsRequest.branch', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1772,
  serialized_end=1854,
)


_FINDMISSINGCOMMITSREPLY = _descriptor.Descriptor(
  name='FindMissingCommitsReply',
  full_name='hangar.FindMissingCommitsReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commits', full_name='hangar.FindMissingCommitsReply.commits', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='branch', full_name='hangar.FindMissingCommitsReply.branch', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingCommitsReply.error', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1856,
  serialized_end=1971,
)


_FINDMISSINGHASHRECORDSREQUEST = _descriptor.Descriptor(
  name='FindMissingHashRecordsRequest',
  full_name='hangar.FindMissingHashRecordsRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingHashRecordsRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hashs', full_name='hangar.FindMissingHashRecordsRequest.hashs', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FindMissingHashRecordsRequest.total_byte_size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1973,
  serialized_end=2060,
)


_FINDMISSINGHASHRECORDSREPLY = _descriptor.Descriptor(
  name='FindMissingHashRecordsReply',
  full_name='hangar.FindMissingHashRecordsReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingHashRecordsReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hashs', full_name='hangar.FindMissingHashRecordsReply.hashs', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='total_byte_size', full_name='hangar.FindMissingHashRecordsReply.total_byte_size', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingHashRecordsReply.error', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2062,
  serialized_end=2182,
)


_FINDMISSINGSCHEMASREQUEST = _descriptor.Descriptor(
  name='FindMissingSchemasRequest',
  full_name='hangar.FindMissingSchemasRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingSchemasRequest.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema_digests', full_name='hangar.FindMissingSchemasRequest.schema_digests', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2184,
  serialized_end=2251,
)


_FINDMISSINGSCHEMASREPLY = _descriptor.Descriptor(
  name='FindMissingSchemasReply',
  full_name='hangar.FindMissingSchemasReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='commit', full_name='hangar.FindMissingSchemasReply.commit', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='schema_digests', full_name='hangar.FindMissingSchemasReply.schema_digests', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error', full_name='hangar.FindMissingSchemasReply.error', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=2253,
  serialized_end=2353,
)

_GETCLIENTCONFIGREPLY_CONFIGENTRY.containing_type = _GETCLIENTCONFIGREPLY
_GETCLIENTCONFIGREPLY.fields_by_name['config'].message_type = _GETCLIENTCONFIGREPLY_CONFIGENTRY
_GETCLIENTCONFIGREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHBRANCHRECORDREQUEST.fields_by_name['rec'].message_type = _BRANCHRECORD
_FETCHBRANCHRECORDREPLY.fields_by_name['rec'].message_type = _BRANCHRECORD
_FETCHBRANCHRECORDREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHDATAREQUEST.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHDATAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHCOMMITREPLY.fields_by_name['record'].message_type = _COMMITRECORD
_FETCHCOMMITREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FETCHSCHEMAREQUEST.fields_by_name['rec'].message_type = _SCHEMARECORD
_FETCHSCHEMAREPLY.fields_by_name['rec'].message_type = _SCHEMARECORD
_FETCHSCHEMAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHBRANCHRECORDREQUEST.fields_by_name['rec'].message_type = _BRANCHRECORD
_PUSHBRANCHRECORDREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHDATAREQUEST.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHDATAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHCOMMITREQUEST.fields_by_name['record'].message_type = _COMMITRECORD
_PUSHCOMMITREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_PUSHSCHEMAREQUEST.fields_by_name['rec'].message_type = _SCHEMARECORD
_PUSHSCHEMAREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGCOMMITSREQUEST.fields_by_name['branch'].message_type = _BRANCHRECORD
_FINDMISSINGCOMMITSREPLY.fields_by_name['branch'].message_type = _BRANCHRECORD
_FINDMISSINGCOMMITSREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGHASHRECORDSREPLY.fields_by_name['error'].message_type = _ERRORPROTO
_FINDMISSINGSCHEMASREPLY.fields_by_name['error'].message_type = _ERRORPROTO
DESCRIPTOR.message_types_by_name['ErrorProto'] = _ERRORPROTO
DESCRIPTOR.message_types_by_name['BranchRecord'] = _BRANCHRECORD
DESCRIPTOR.message_types_by_name['HashRecord'] = _HASHRECORD
DESCRIPTOR.message_types_by_name['CommitRecord'] = _COMMITRECORD
DESCRIPTOR.message_types_by_name['SchemaRecord'] = _SCHEMARECORD
DESCRIPTOR.message_types_by_name['PingRequest'] = _PINGREQUEST
DESCRIPTOR.message_types_by_name['PingReply'] = _PINGREPLY
DESCRIPTOR.message_types_by_name['GetClientConfigRequest'] = _GETCLIENTCONFIGREQUEST
DESCRIPTOR.message_types_by_name['GetClientConfigReply'] = _GETCLIENTCONFIGREPLY
DESCRIPTOR.message_types_by_name['FetchBranchRecordRequest'] = _FETCHBRANCHRECORDREQUEST
DESCRIPTOR.message_types_by_name['FetchBranchRecordReply'] = _FETCHBRANCHRECORDREPLY
DESCRIPTOR.message_types_by_name['FetchDataRequest'] = _FETCHDATAREQUEST
DESCRIPTOR.message_types_by_name['FetchDataReply'] = _FETCHDATAREPLY
DESCRIPTOR.message_types_by_name['FetchCommitRequest'] = _FETCHCOMMITREQUEST
DESCRIPTOR.message_types_by_name['FetchCommitReply'] = _FETCHCOMMITREPLY
DESCRIPTOR.message_types_by_name['FetchSchemaRequest'] = _FETCHSCHEMAREQUEST
DESCRIPTOR.message_types_by_name['FetchSchemaReply'] = _FETCHSCHEMAREPLY
DESCRIPTOR.message_types_by_name['PushBranchRecordRequest'] = _PUSHBRANCHRECORDREQUEST
DESCRIPTOR.message_types_by_name['PushBranchRecordReply'] = _PUSHBRANCHRECORDREPLY
DESCRIPTOR.message_types_by_name['PushDataRequest'] = _PUSHDATAREQUEST
DESCRIPTOR.message_types_by_name['PushDataReply'] = _PUSHDATAREPLY
DESCRIPTOR.message_types_by_name['PushCommitRequest'] = _PUSHCOMMITREQUEST
DESCRIPTOR.message_types_by_name['PushCommitReply'] = _PUSHCOMMITREPLY
DESCRIPTOR.message_types_by_name['PushSchemaRequest'] = _PUSHSCHEMAREQUEST
DESCRIPTOR.message_types_by_name['PushSchemaReply'] = _PUSHSCHEMAREPLY
DESCRIPTOR.message_types_by_name['FindMissingCommitsRequest'] = _FINDMISSINGCOMMITSREQUEST
DESCRIPTOR.message_types_by_name['FindMissingCommitsReply'] = _FINDMISSINGCOMMITSREPLY
DESCRIPTOR.message_types_by_name['FindMissingHashRecordsRequest'] = _FINDMISSINGHASHRECORDSREQUEST
DESCRIPTOR.message_types_by_name['FindMissingHashRecordsReply'] = _FINDMISSINGHASHRECORDSREPLY
DESCRIPTOR.message_types_by_name['FindMissingSchemasRequest'] = _FINDMISSINGSCHEMASREQUEST
DESCRIPTOR.message_types_by_name['FindMissingSchemasReply'] = _FINDMISSINGSCHEMASREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ErrorProto = _reflection.GeneratedProtocolMessageType('ErrorProto', (_message.Message,), {
  'DESCRIPTOR' : _ERRORPROTO,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.ErrorProto)
  })
_sym_db.RegisterMessage(ErrorProto)

BranchRecord = _reflection.GeneratedProtocolMessageType('BranchRecord', (_message.Message,), {
  'DESCRIPTOR' : _BRANCHRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.BranchRecord)
  })
_sym_db.RegisterMessage(BranchRecord)

HashRecord = _reflection.GeneratedProtocolMessageType('HashRecord', (_message.Message,), {
  'DESCRIPTOR' : _HASHRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.HashRecord)
  })
_sym_db.RegisterMessage(HashRecord)

CommitRecord = _reflection.GeneratedProtocolMessageType('CommitRecord', (_message.Message,), {
  'DESCRIPTOR' : _COMMITRECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.CommitRecord)
  })
_sym_db.RegisterMessage(CommitRecord)

SchemaRecord = _reflection.GeneratedProtocolMessageType('SchemaRecord', (_message.Message,), {
  'DESCRIPTOR' : _SCHEMARECORD,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.SchemaRecord)
  })
_sym_db.RegisterMessage(SchemaRecord)

PingRequest = _reflection.GeneratedProtocolMessageType('PingRequest', (_message.Message,), {
  'DESCRIPTOR' : _PINGREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PingRequest)
  })
_sym_db.RegisterMessage(PingRequest)

PingReply = _reflection.GeneratedProtocolMessageType('PingReply', (_message.Message,), {
  'DESCRIPTOR' : _PINGREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PingReply)
  })
_sym_db.RegisterMessage(PingReply)

GetClientConfigRequest = _reflection.GeneratedProtocolMessageType('GetClientConfigRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETCLIENTCONFIGREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.GetClientConfigRequest)
  })
_sym_db.RegisterMessage(GetClientConfigRequest)

GetClientConfigReply = _reflection.GeneratedProtocolMessageType('GetClientConfigReply', (_message.Message,), {

  'ConfigEntry' : _reflection.GeneratedProtocolMessageType('ConfigEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETCLIENTCONFIGREPLY_CONFIGENTRY,
    '__module__' : 'hangar_service_pb2'
    # @@protoc_insertion_point(class_scope:hangar.GetClientConfigReply.ConfigEntry)
    })
  ,
  'DESCRIPTOR' : _GETCLIENTCONFIGREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.GetClientConfigReply)
  })
_sym_db.RegisterMessage(GetClientConfigReply)
_sym_db.RegisterMessage(GetClientConfigReply.ConfigEntry)

FetchBranchRecordRequest = _reflection.GeneratedProtocolMessageType('FetchBranchRecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHBRANCHRECORDREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchBranchRecordRequest)
  })
_sym_db.RegisterMessage(FetchBranchRecordRequest)

FetchBranchRecordReply = _reflection.GeneratedProtocolMessageType('FetchBranchRecordReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHBRANCHRECORDREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchBranchRecordReply)
  })
_sym_db.RegisterMessage(FetchBranchRecordReply)

FetchDataRequest = _reflection.GeneratedProtocolMessageType('FetchDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHDATAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchDataRequest)
  })
_sym_db.RegisterMessage(FetchDataRequest)

FetchDataReply = _reflection.GeneratedProtocolMessageType('FetchDataReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHDATAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchDataReply)
  })
_sym_db.RegisterMessage(FetchDataReply)

FetchCommitRequest = _reflection.GeneratedProtocolMessageType('FetchCommitRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHCOMMITREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchCommitRequest)
  })
_sym_db.RegisterMessage(FetchCommitRequest)

FetchCommitReply = _reflection.GeneratedProtocolMessageType('FetchCommitReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHCOMMITREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchCommitReply)
  })
_sym_db.RegisterMessage(FetchCommitReply)

FetchSchemaRequest = _reflection.GeneratedProtocolMessageType('FetchSchemaRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHSCHEMAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchSchemaRequest)
  })
_sym_db.RegisterMessage(FetchSchemaRequest)

FetchSchemaReply = _reflection.GeneratedProtocolMessageType('FetchSchemaReply', (_message.Message,), {
  'DESCRIPTOR' : _FETCHSCHEMAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FetchSchemaReply)
  })
_sym_db.RegisterMessage(FetchSchemaReply)

PushBranchRecordRequest = _reflection.GeneratedProtocolMessageType('PushBranchRecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHBRANCHRECORDREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushBranchRecordRequest)
  })
_sym_db.RegisterMessage(PushBranchRecordRequest)

PushBranchRecordReply = _reflection.GeneratedProtocolMessageType('PushBranchRecordReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHBRANCHRECORDREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushBranchRecordReply)
  })
_sym_db.RegisterMessage(PushBranchRecordReply)

PushDataRequest = _reflection.GeneratedProtocolMessageType('PushDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHDATAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushDataRequest)
  })
_sym_db.RegisterMessage(PushDataRequest)

PushDataReply = _reflection.GeneratedProtocolMessageType('PushDataReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHDATAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushDataReply)
  })
_sym_db.RegisterMessage(PushDataReply)

PushCommitRequest = _reflection.GeneratedProtocolMessageType('PushCommitRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHCOMMITREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushCommitRequest)
  })
_sym_db.RegisterMessage(PushCommitRequest)

PushCommitReply = _reflection.GeneratedProtocolMessageType('PushCommitReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHCOMMITREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushCommitReply)
  })
_sym_db.RegisterMessage(PushCommitReply)

PushSchemaRequest = _reflection.GeneratedProtocolMessageType('PushSchemaRequest', (_message.Message,), {
  'DESCRIPTOR' : _PUSHSCHEMAREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushSchemaRequest)
  })
_sym_db.RegisterMessage(PushSchemaRequest)

PushSchemaReply = _reflection.GeneratedProtocolMessageType('PushSchemaReply', (_message.Message,), {
  'DESCRIPTOR' : _PUSHSCHEMAREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.PushSchemaReply)
  })
_sym_db.RegisterMessage(PushSchemaReply)

FindMissingCommitsRequest = _reflection.GeneratedProtocolMessageType('FindMissingCommitsRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGCOMMITSREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingCommitsRequest)
  })
_sym_db.RegisterMessage(FindMissingCommitsRequest)

FindMissingCommitsReply = _reflection.GeneratedProtocolMessageType('FindMissingCommitsReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGCOMMITSREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingCommitsReply)
  })
_sym_db.RegisterMessage(FindMissingCommitsReply)

FindMissingHashRecordsRequest = _reflection.GeneratedProtocolMessageType('FindMissingHashRecordsRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGHASHRECORDSREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingHashRecordsRequest)
  })
_sym_db.RegisterMessage(FindMissingHashRecordsRequest)

FindMissingHashRecordsReply = _reflection.GeneratedProtocolMessageType('FindMissingHashRecordsReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGHASHRECORDSREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingHashRecordsReply)
  })
_sym_db.RegisterMessage(FindMissingHashRecordsReply)

FindMissingSchemasRequest = _reflection.GeneratedProtocolMessageType('FindMissingSchemasRequest', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGSCHEMASREQUEST,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingSchemasRequest)
  })
_sym_db.RegisterMessage(FindMissingSchemasRequest)

FindMissingSchemasReply = _reflection.GeneratedProtocolMessageType('FindMissingSchemasReply', (_message.Message,), {
  'DESCRIPTOR' : _FINDMISSINGSCHEMASREPLY,
  '__module__' : 'hangar_service_pb2'
  # @@protoc_insertion_point(class_scope:hangar.FindMissingSchemasReply)
  })
_sym_db.RegisterMessage(FindMissingSchemasReply)


DESCRIPTOR._options = None
_GETCLIENTCONFIGREPLY_CONFIGENTRY._options = None

_HANGARSERVICE = _descriptor.ServiceDescriptor(
  name='HangarService',
  full_name='hangar.HangarService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=2356,
  serialized_end=3705,
  methods=[
  _descriptor.MethodDescriptor(
    name='PING',
    full_name='hangar.HangarService.PING',
    index=0,
    containing_service=None,
    input_type=_PINGREQUEST,
    output_type=_PINGREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetClientConfig',
    full_name='hangar.HangarService.GetClientConfig',
    index=1,
    containing_service=None,
    input_type=_GETCLIENTCONFIGREQUEST,
    output_type=_GETCLIENTCONFIGREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchBranchRecord',
    full_name='hangar.HangarService.FetchBranchRecord',
    index=2,
    containing_service=None,
    input_type=_FETCHBRANCHRECORDREQUEST,
    output_type=_FETCHBRANCHRECORDREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchData',
    full_name='hangar.HangarService.FetchData',
    index=3,
    containing_service=None,
    input_type=_FETCHDATAREQUEST,
    output_type=_FETCHDATAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchCommit',
    full_name='hangar.HangarService.FetchCommit',
    index=4,
    containing_service=None,
    input_type=_FETCHCOMMITREQUEST,
    output_type=_FETCHCOMMITREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchSchema',
    full_name='hangar.HangarService.FetchSchema',
    index=5,
    containing_service=None,
    input_type=_FETCHSCHEMAREQUEST,
    output_type=_FETCHSCHEMAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushBranchRecord',
    full_name='hangar.HangarService.PushBranchRecord',
    index=6,
    containing_service=None,
    input_type=_PUSHBRANCHRECORDREQUEST,
    output_type=_PUSHBRANCHRECORDREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushData',
    full_name='hangar.HangarService.PushData',
    index=7,
    containing_service=None,
    input_type=_PUSHDATAREQUEST,
    output_type=_PUSHDATAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushCommit',
    full_name='hangar.HangarService.PushCommit',
    index=8,
    containing_service=None,
    input_type=_PUSHCOMMITREQUEST,
    output_type=_PUSHCOMMITREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushSchema',
    full_name='hangar.HangarService.PushSchema',
    index=9,
    containing_service=None,
    input_type=_PUSHSCHEMAREQUEST,
    output_type=_PUSHSCHEMAREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingCommits',
    full_name='hangar.HangarService.FetchFindMissingCommits',
    index=10,
    containing_service=None,
    input_type=_FINDMISSINGCOMMITSREQUEST,
    output_type=_FINDMISSINGCOMMITSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingHashRecords',
    full_name='hangar.HangarService.FetchFindMissingHashRecords',
    index=11,
    containing_service=None,
    input_type=_FINDMISSINGHASHRECORDSREQUEST,
    output_type=_FINDMISSINGHASHRECORDSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FetchFindMissingSchemas',
    full_name='hangar.HangarService.FetchFindMissingSchemas',
    index=12,
    containing_service=None,
    input_type=_FINDMISSINGSCHEMASREQUEST,
    output_type=_FINDMISSINGSCHEMASREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingCommits',
    full_name='hangar.HangarService.PushFindMissingCommits',
    index=13,
    containing_service=None,
    input_type=_FINDMISSINGCOMMITSREQUEST,
    output_type=_FINDMISSINGCOMMITSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingHashRecords',
    full_name='hangar.HangarService.PushFindMissingHashRecords',
    index=14,
    containing_service=None,
    input_type=_FINDMISSINGHASHRECORDSREQUEST,
    output_type=_FINDMISSINGHASHRECORDSREPLY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='PushFindMissingSchemas',
    full_name='hangar.HangarService.PushFindMissingSchemas',
    index=15,
    containing_service=None,
    input_type=_FINDMISSINGSCHEMASREQUEST,
    output_type=_FINDMISSINGSCHEMASREPLY,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_HANGARSERVICE)

DESCRIPTOR.services_by_name['HangarService'] = _HANGARSERVICE

# @@protoc_insertion_point(module_scope)
