print("import: 'hangar'")
import hangar

print("import: 'hangar.diagnostics'")
import hangar.diagnostics

print("import: 'hangar.records'")
import hangar.records

print("import: 'hangar.remote'")
import hangar.remote

print("import: 'hangar.backends'")
import hangar.backends

print("import: 'hangar.columns'")
import hangar.columns

